#!/bin/bash

readonly scriptdir=`dirname $(readlink -f $0)`

readonly solversdir="$scriptdir/../solvers/"

readonly cvc4_base="$solversdir/CVC4-base/bin/run"
readonly cvc4_ext="$solversdir/CVC4-ext/bin/run"
readonly z3="$solversdir/Z3/bin/run"

readonly benchmarks_all="$scriptdir/benchmark-set-FP-all"
readonly benchmarks_sel="$scriptdir/benchmark-set-FP-25"

benchmarks_file="$benchmarks_sel"

tmpdir="$scriptdir/paper-results"

readonly runsdir="$scriptdir/results-eval"

readonly timelimit=1800
readonly spacelimit=8000

readonly runlim="$scriptdir/../scripts/runlim"
readonly runlim_opts="--time-limit=$timelimit --space-limit=$spacelimit"

readonly compare_script="$scriptdir/../scripts/compare-smt-runs.py"

function die
{
  echo "** $(basename $0): $1"
  exit 1
}

function msg
{
  echo "[CAV19] $1"
}

function run_solver()
{
  msg "$1.. Running $2 on `basename $benchmarks_file`"
  workdir=$4
  rm -rf $workdir
  mkdir -p $workdir
  solver="$3"
  benchmarks_file=$5
  for f in `cat $benchmarks_file`
  do
    out="$(echo $f | sed 's/^benchmarks\///' | sed -e 's/benchmarks-25\///' | sed 's/FP\///' | sed 's/FP-25\///' | sed -e 's,/,-,g')"
    log="$workdir/${out}.log"
    err="$workdir/${out}.err"
    echo -n "c launch: " > $log
    echo -n "$benchmarks_file $workdir $runlim_opts " >> $log
    echo -ne "$solver $opts\n" >> $log
    echo "c benchmark: $f" >> $log
    echo "c solver: $solver" >> $log
    echo "c solver options: $opts" >> $log
    $runlim $runlim_opts $solver $opts $f >> $log 2> $err
  done
}

function show_stats()
{
  workdir=$3
  rm -rf $tmpdir
  a=unknown
  while [[ $a != "y" && $a != "n" ]]
  do
    msg
    msg "$1 .. Show statistics for $2: y/n"
    read a
    if [[ $a == "n" ]]; then break; fi
    if [[ $a == "y" ]]
    then
      msg
      msg "Compare current run '`basename $workdir`' with provided log files from paper ('`basename $tmpdir`')..."
      msg
      if [ -d "$tmpdir" -o -f "$tmpdir" ]; then
        die "tmp dir '$tmpdir' does already exist"
      fi
      mkdir -p $tmpdir
      for f in $workdir/*
      do
        cp "$runsdir/$2/`basename $f`" $tmpdir
      done
      $compare_script --no-colors -s time_cpu $workdir $tmpdir
      rm -rf $tmpdir
      break
    fi
    done
}

while [ $# -gt 0 ]
do
  case $1 in
    -h|--help)
      echo -n "usage: $(basename $0)"
      echo -e " [ -h | --help | -a ] [<benchmarks-file>]"
      echo -n " -a        Run all benchmarks. Default is to run a "
      echo -e "selection of 50 benchmarks."
      exit 0 ;;
    -a) benchmarks_file="$benchmarks_all" ;;
    -*|--*)
      die "invalid option, try '-h'" ;;
    *) break ;;
  esac
  shift
done

if [ $# -gt 0 ]; then
  die "invalid number of arguments, try '-h'"
fi

# check if binaries are available

if [ ! -f "$cvc4_base" ]
then
  die "could not find CVC4-base binary (not built?)"
fi
if [ ! -f "$cvc4_ext" ]
then
  die "could not find CVC4-ext binary (not built?)"
fi
if [ ! -f "$z3" ]
then
  die "could not find Z3 binary (not built?)"
fi

if [ ! -f "$benchmarks_sel" ]
then
  die "could not find benchmarks (selection) file"
fi

workdir_cvc4_base=$scriptdir/`date +%Y-%m-%d-``basename $benchmarks_file`-cvc4_base
workdir_cvc4_ext=$scriptdir/`date +%Y-%m-%d-``basename $benchmarks_file`-cvc4_ext
workdir_z3=$scriptdir/`date +%Y-%m-%d-``basename $benchmarks_file`-z3


# create working directories (for log files and other output)
# and run experiments


msg "\"Invertibility Conditions for Floating-Point Formulas\""
msg "Martin Brain, Aina Niemetz, Mathias Preiner, Andrew Reynolds,"
msg "Clark Barrett and Cesare Tinelli"
msg
msg "Reproducing results from Section 5 from provided log files in '$runsdir'..."
msg
msg "totals"
msg
$compare_script --no-colors -g -s g_solved $runsdir/CVC4-base $runsdir/CVC4-ext $runsdir/Z3
msg
msg "unique instances: CVC4-base vs CVC4-ext"
msg
$compare_script --no-colors -g -s g_solved -u $runsdir/CVC4-base $runsdir/CVC4-ext

msg
msg "Running the experiments ..."
msg 

run_solver 1 CVC4-base $cvc4_base $workdir_cvc4_base $benchmarks_file
run_solver 2 CVC4-ext $cvc4_ext $workdir_cvc4_ext $benchmarks_file
run_solver 3 Z3 $z3 $workdir_z3 $benchmarks_file

show_stats 1 CVC4-base $workdir_cvc4_base
show_stats 2 CVC4-ext $workdir_cvc4_ext
show_stats 3 Z3 $workdir_z3
