#! /usr/bin/env python3


from argparse import ArgumentParser, REMAINDER, ArgumentDefaultsHelpFormatter
import os
import sys
import re

class CmpSMTException (Exception):
    def __init__ (self, msg):
        self.msg = msg
    def __str__ (self):
        return "[cmpsmt] Error: {0:s}".format(self.msg)

# required for loading additional resources for --html
SCRIPTDIR = os.path.dirname(os.path.realpath(__file__))

g_args = None
g_benchmarks = set()

CSV_DELIM = ','

COLOR_BEST = '\033[36m'
COLOR_DIFF = '\033[32m'
COLOR_DISC = '\033[31m'
COLOR_STAT = '\033[33m'
COLOR_NOCOLOR = '\033[0m'

HTML_CLASS = {
    COLOR_BEST: 'best',
    COLOR_DIFF: 'diff',
    COLOR_DISC: 'disc',
    COLOR_STAT: 'stat'
}

RESULT_SAT = 10
RESULT_UNSAT = 20
RESULT_UNKNOWN = 30
RESULT_SYNTH = 40

STATUS_ERR = 'ee'
STATUS_OK = 'ok'
STATUS_MEM = 'mo'
STATUS_TIME = 'to'
STATUS_DIS = 'di'

def _get_name_and_ext (fname):
    return ("".join(fname.rpartition('.')[:-2]), fname.rpartition('.')[-1])

def _is_float(s):
    return s.lstrip('-').replace('.', '', 1).isdigit()

def _is_int(s):
    if isinstance(s, int):
        return True
    return s.lstrip('-').isdigit()

def _is_number(s):
    return _is_float(s)

def _is_tptp_result_unknown(x):
  return 'SZS status GaveUp' in x

def _is_tptp_result_unsat(x):
  return 'SZS status Theorem' in x or 'SZS status ContradictoryAxioms' in x or \
    'SZS status Unsatisfiable' in x

def _is_tptp_result_sat(x):
  return 'SZS status CounterSatisfiable' in x or 'SZS status Satisfiable' in x

def _is_tptp_result(x):
  return _is_tptp_result_unknown(x) or _is_tptp_result_unsat(x) or \
    _is_tptp_result_sat(x)

def _is_synth_result(x):
  return x.startswith('(define-fun') or x.startswith('Synthesis time')

def _is_smt_result(x):
  return x in ['sat', 'unsat', 'unknown']

def _is_solver_result(x):
    if g_args.synth:
        return x == 'unknown' or _is_synth_result(x)
    else:
        return _is_smt_result(x) or (g_args.tptp and _is_tptp_result(x))


def _result_to_num(s):
    if s == 'sat' or (g_args.tptp and _is_tptp_result_sat(s)):
        return RESULT_SAT
    elif not g_args.synth and s == 'unsat' or \
      (g_args.tptp and _is_tptp_result_unsat(s)):
        return RESULT_UNSAT
    elif s == 'unknown' or (g_args.tptp and _is_tptp_result_unknown(s)):
        return RESULT_UNKNOWN
    elif g_args.synth and _is_synth_result(s):
        return RESULT_SYNTH
    return 0

# per directory/file flag
# column_name : <colname>, <keyword>, <filter>
FILTER_LOG = [
  ('result',
   'RES',
   lambda x: _is_solver_result(x),
   lambda x: _result_to_num(x)),
  ('benchmark',
   'BENCH',
   lambda x: 'benchmark:' in x,
   lambda x: x.split()[2]),
]

def _read_status(line):
    status = line.split(':')[1].strip()
    if 'ok' in status:
        return STATUS_OK
    # if process gets killed by runlim it sometimes only returns
    # signal(9) instead of time
    elif 'time' in status or 'signal(9)' in status:
        return STATUS_TIME
    elif 'memory' in status:
        return STATUS_MEM
    return STATUS_ERR


# column_name : <colname>, <keyword>, <filter>, <format>
FILTER_ERR = [
  ('status',
   'Stat',
   lambda x: '[runlim] status:' in x,
   _read_status),
  ('exit_code',
   'Exit',
   lambda x: '[runlim] result:' in x,
   lambda x: x.split()[2]),
  ('time_real',
   'Real[s]',
   lambda x: '[runlim] real:' in x,
   lambda x: x.split()[2]),
  ('time_cpu',
   'Cpu[s]',
   lambda x: '[runlim] time:' in x,
   lambda x: x.split()[2]),
  ('memory',
   'Mem[MB]',
   lambda x: '[runlim] space:' in x,
   lambda x: x.split()[2]),
]

# these columns will be computed in _normalize_data
GROUP_COLUMNS = {
  'g_solved':  'Slvd',
  'g_total':   'Tot',
  'g_time':    'To',
  'g_mem':     'Mo',
  'g_err':     'Err',
  'g_sat':     'Sat',
  'g_unsat':   'Unsat',
  'g_unknown': 'Unk',
  'g_uniq':    'Uniq',
}

TOTALS_KEY = '__totals'

filter_log_keys = set([t[0] for t in FILTER_LOG])
filter_err_keys = set([t[0] for t in FILTER_ERR])

filter_log_dict = dict([(t[0], t[1:]) for t in FILTER_LOG])
filter_err_dict = dict([(t[0], t[1:]) for t in FILTER_ERR])

assert(filter_log_keys.isdisjoint(filter_err_keys))

STATS_KEYS = [t[0] for t in FILTER_LOG]
STATS_KEYS.extend([t[0] for t in FILTER_ERR])
STATS_KEYS.extend(GROUP_COLUMNS.keys())

g_file_stats = {}
g_total_stats = {}

def _format_status(l):
    if STATUS_DIS in l:
        return STATUS_DIS
    if STATUS_ERR in l:
        return STATUS_ERR
    elif STATUS_OK in l:
        return STATUS_OK
    return '/'.join(set(l))

g_format_column_sum = {
    'status' : _format_status,
}

def _cast(s):
    if _is_int(s):
        return int(s)
    elif _is_float(s):
        return float(s)
    return s

def _filter_data(d, file, filters):
    global g_file_stats

    (f_name, f_ext) = _get_name_and_ext(file)

    for t in filters:
        k = t[0]
        if k not in g_file_stats:
            g_file_stats[k] = {}
        if d not in g_file_stats[k]:
            g_file_stats[k][d] = {}
        if f_name not in g_file_stats[k][d]:
            g_file_stats[k][d][f_name] = None

    if os.stat(os.path.join(d, file)).st_size == 0:
        return

    used_filters = set()
    with open(os.path.join(d, file), 'r') as infile:
        lines = infile.readlines()
        for line in lines:
            line = line.strip()

            if len(used_filters) == len(filters):
                break

            for t in filters:
                assert(len(t) == 4)

                k = t[0]
                # value already extracted for current file
                if k in used_filters:
                    continue

                f_match = t[2]
                f_val = t[3]

                if not f_match(line):
                    continue

                val = _cast(f_val(line))

                if not g_args.incremental or k != 'result':
                    used_filters.add(k)

                # CHECK: *.out are model files and may be obsolete
                if f_ext == 'out' \
                   and line in ('sat', 'unsat', 'unknown'):
                       continue

                if not g_args.incremental or k != 'result':
                    assert(g_file_stats[k][d][f_name] == None)
                    g_file_stats[k][d][f_name] = val
                else:
                    if g_file_stats[k][d][f_name] is None:
                        g_file_stats[k][d][f_name] = []
                    g_file_stats[k][d][f_name].append(val)



def _read_log_file(d, f):
    _filter_data(d, f, FILTER_LOG)


def _read_err_file(d, f):
    global g_file_stats

    try:
        _filter_data(d, f, FILTER_ERR)
    except CmpSMTException as e:
        raise CmpSMTException("{} in file {}".format(str(e), f))

def _init_missing_files(data):
    global g_benchmarks

    for k in data:
        for d in g_args.dirs:
            for f in g_benchmarks:
                if d not in data[k]:
                    data[k][d] = {}
                if f not in data[k][d]:
                    data[k][d][f] = None

def _normalize_data(data):
    global g_args

    assert 'result' in data

    # reset timeout if given
    if g_args.timeout:
        for d in data['time_cpu']:
            for f in data['time_cpu'][d]:
                if data['time_cpu'][d][f] is None or \
                   data['time_cpu'][d][f] > g_args.timeout[d]:
                       data['time_cpu'][d][f] = g_args.timeout[d]
                       data['status'][d][f] = STATUS_TIME
                       data['result'][d][f] = 1


    # initialize group columns derived from status
    for k in ['g_solved', 'g_total', 'g_time', 'g_mem', 'g_err',
              'g_sat','g_unsat', 'g_unknown']:
        assert(k not in data)
        data[k] = {}
        for d in data['status']:
            assert(d not in data[k])
            data[k][d] = {}
            for f in data['status'][d]:
                assert(f not in data[k][d])
                if k == 'g_total':
                    data[k][d][f] = 1
                else:
                    data[k][d][f] = 0

    # check for discrepancies
    for f in g_benchmarks:
        results = _get_values(data, f, 'result')
        found_dis = False
        if g_args.incremental:
            # counter for each results list
            idx = [0 for r in results]
            done = False
            # check all incremental results line by line
            while not done:
                done = True
                r = set()
                for i in range(len(results)):
                    cur_idx = idx[i]
                    cur_res = results[i]
                    if cur_res and cur_idx < len(cur_res):
                        r.add(cur_res[cur_idx])
                        idx[i] += 1
                        done = False
                if RESULT_SAT in r and RESULT_UNSAT in r:
                    found_dis = True
                    break

        elif RESULT_SAT in results and RESULT_UNSAT in results:
            found_dis = True

        if found_dis and not g_args.ignore_discrepancy:
            for d in g_args.dirs:
                data['status'][d][f] = STATUS_DIS

    # compute values for group columns derived from status
    for d in data['status']:
        for f in data['status'][d]:
            s = data['status'][d][f]
            r = data['result'][d][f]

            if g_args.incremental:
                if s == STATUS_OK:
                    data['g_solved'][d][f] = 1
                if data['result'][d][f] is not None:
                    data['g_sat'][d][f] = \
                            data['result'][d][f].count(RESULT_SAT)
                    data['g_unsat'][d][f] = \
                            data['result'][d][f].count(RESULT_UNSAT)
                    data['g_solved'][d][f] = \
                            data['g_sat'][d][f] + data['g_unsat'][d][f]
            elif s == STATUS_OK and r in (RESULT_SAT, RESULT_UNSAT):
                data['g_solved'][d][f] = 1
                if r == RESULT_SAT:
                    data['g_sat'][d][f] = 1
                else:
                    assert(r == RESULT_UNSAT)
                    data['g_unsat'][d][f] = 1
            elif g_args.synth and s == STATUS_OK and r == RESULT_SYNTH:
                data['g_solved'][d][f] = 1
            elif s == STATUS_OK and r == RESULT_UNKNOWN:
                data['g_unknown'][d][f] = 1
            elif s == STATUS_TIME:
                data['g_time'][d][f] = 1
            elif s == STATUS_MEM:
                data['g_mem'][d][f] = 1
                if g_args.pen:
                    data['time_cpu'][d][f] = g_args.pen
            elif s != STATUS_DIS:
                data['status'][d][f] = STATUS_ERR
                data['g_err'][d][f] = 1
                if g_args.pen:
                    data['time_cpu'][d][f] = g_args.pen

    # collect data for virtual best solver
    if g_args.virtual_best:
        if g_args.virtual_best_pp:
            bdir = g_args.dirs[0]   # base run
            prep_dirs = g_args.dirs[1:] # "preprocessing" runs

            for pdir in prep_dirs:
                vbpdir = "{} + {}".format(_base_dir(bdir), _base_dir(pdir))

                # initialize for virtual best solver
                for k in data.keys():
                    if vbpdir not in data[k]:
                        data[k][vbpdir] = {}

                for f in g_benchmarks:
                    if data["time_cpu"][bdir][f] < g_args.timeout[bdir]:
                        time_bdir = data["time_cpu"][bdir][f]
                    else:
                        time_bdir = g_args.timeout[bdir]

                    if data["time_cpu"][pdir][f] < g_args.timeout[pdir]:
                        time_pdir = data["time_cpu"][pdir][f]
                    else:
                        time_pdir = g_args.timeout[pdir]

                    if data["status"][pdir][f] in ["ok", "time"] \
                       and time_pdir < g_args.timeout[pdir]:
                        for k in data.keys():
                            data[k][vbpdir][f] = data[k][pdir][f]
                    elif data["status"][bdir][f] == "ok":
                        for k in data.keys():
                            data[k][vbpdir][f] = data[k][bdir][f]
                        time_vbp = round(time_bdir + g_args.timeout[pdir], 2)
                        data["time_cpu"][vbpdir][f] = time_vbp
                        if time_vbp >= g_args.timeout[bdir]:
                            data["status"][vbpdir][f] = "time"
                            data["time_cpu"][vbpdir][f] = g_args.timeout[bdir]
                            if not g_args.grouped:
                                data["result"][vbpdir][f] = 1
                    else:
                        for k in data.keys():
                            data[k][vbpdir][f] = data[k][bdir][f]
                g_args.dirs.append(vbpdir)
        else:
            vb_dir = "virtual best solver (portfolio)"

            for f in g_benchmarks:
                v = []
                for d in g_args.dirs:
                    if data['time_cpu'][d][f] is not None \
                        and data['status'][d][f] == STATUS_OK \
                        and data['result'][d][f] in \
                                [RESULT_SAT, RESULT_UNSAT, RESULT_SYNTH]:
                        v.append((data['time_cpu'][d][f], d))
                v = sorted(v)

                best_dir = None
                if len(v) > 0:
                    best_dir = v[0][1]
                for k in data.keys():
                    if vb_dir not in data[k]:
                        data[k][vb_dir] = {}
                    if best_dir is None:
                        data[k][vb_dir][f] = None
                    else:
                        data[k][vb_dir][f] = data[k][best_dir][f]
            g_args.dirs.append(vb_dir)


    # add uniquely solved column
    if g_args.unique:
        data['g_uniq'] = {}
        for f in g_benchmarks:
            stats = []
            for d in data['status']:
                if d not in data['g_uniq']:
                    data['g_uniq'][d] = {}
                data['g_uniq'][d][f] = 0
                if data['status'][d][f] == STATUS_OK \
                    and data['result'][d][f] in \
                        [RESULT_SAT, RESULT_UNSAT, RESULT_SYNTH]:
                    stats.append(d)

            if len(stats) == 1:
                data['g_uniq'][stats[0]][f] = 1

def _save_cache_file(dir):
    global g_file_stats

    sum_filename = "{}/cache".format(dir)
    keys = sorted(g_file_stats.keys())
    if len(keys) == 0:
        return
    files = []
    for f in g_file_stats[keys[0]][dir]:
        files.append(f)

    with open(sum_filename, "w") as sum_file:
        sum_file.write("{}\n".format(CSV_DELIM.join(keys)))
        for f in sorted(files):
            data = []
            for k in keys:
                data.append(str(g_file_stats[k][dir][f]))
            sum_file.write("{}{}{}\n".format(
                f, CSV_DELIM, CSV_DELIM.join(data)))

def _read_cache_file(dir):
    global g_file_stats, g_benchmarks

    sum_filename = "{}/cache".format(dir)
    if os.path.isfile(sum_filename):
        with open(sum_filename, "r") as sum_file:
            keys = None
            for line in sum_file:
                # initialize keys
                if keys is None:
                    keys = line.strip().split(CSV_DELIM)

                    expected_keys = list(filter_err_keys)
                    expected_keys.extend(filter_log_keys)

                    missing_keys = set(expected_keys) - set(keys)
                    if len(missing_keys) > 0:
                        print("column(s) {} missing for directory '{}'. "\
                              "extracting data from files".format(
                                  list(missing_keys), dir))
                        return False
                else:
                    cols = line.strip().split(CSV_DELIM)
                    f = cols[0]
                    if f not in g_benchmarks:
                        g_benchmarks.add(f)
                    data = cols[1:]
                    assert(len(data) == len(keys))
                    for i in range(len(keys)):
                        k = keys[i]
                        assert(k in STATS_KEYS)
                        if k not in g_file_stats:
                            g_file_stats[k] = {}
                        if dir not in g_file_stats[k]:
                            g_file_stats[k][dir] = {}
                        if k in filter_log_dict:
                            t = filter_log_dict[k]
                        else:
                            assert(k in filter_err_dict)
                            t = filter_err_dict[k]
                        assert(len(t) == 3)
                        assert(f not in g_file_stats[k][dir])
                        if data[i] == "None":
                            g_file_stats[k][dir][f] = None
                        else:
                            g_file_stats[k][dir][f] = _cast(data[i])
            return True
    return False

def _read_data (dirs):
    for d in dirs:
        # read cached data
        if not g_args.incremental and _read_cache_file(d):
            continue

        # extract data from files
        for root, _, files in os.walk(d):
            assert(root.startswith(d))
            root = root[len(d):].lstrip("/")
            for name in files:
                f = name
                if len(root) > 0:
                    f = "{}/{}".format(root, name)
                (f_name, f_ext) = _get_name_and_ext (f)
                if f_ext == "log":
                    full_log = "{}/{}".format(d, f)
                    full_err = "{}err".format(full_log[:-3])
                    assert(os.path.isfile(full_log))
                    if not os.path.exists(full_err):
                        raise CmpSMTException (
                                "missing .err file for '{}'".format(full_log))
                    # init data
                    if f_name not in g_benchmarks:
                        g_benchmarks.add(f_name)
                    _read_err_file (d, "{}.err".format(f_name))
                    _read_log_file (d, "{}.log".format(f_name))

        # create cache file
        # TODO: seralization does not work with incremental runs yet since
        #       we need to support multiple values for one key
        if not g_args.incremental:
            _save_cache_file(d)

def _pick_data(benchmarks, data):
    global g_args

    sort_reverse = False
    f_cmp = lambda x, y: x * (1 + g_args.dmin) <= y
    if g_args.cmp_col == 'g_solved':
        sort_reverse = True
        f_cmp = lambda x, y: x > y

    best_stats = {}
    diff_stats = {}

    k = g_args.cmp_col
    for f in benchmarks:
        if k == 'g_solved':
            v = sorted(\
                [(data[k][d][f], d) \
                    for d in g_args.dirs if data[k][d][f] is not None],
                reverse=sort_reverse)
        else:
            v = sorted(\
                [(data[k][d][f], d) \
                    for d in g_args.dirs \
                    if data[k][d][f] is not None \
                       and data['result'][d][f] in (RESULT_SAT, RESULT_UNSAT)],
                reverse=sort_reverse)

        # strings are not considered for diff/best values
        if len(v) == 0 or isinstance(v[0][0], str):
            best_stats[f] = None
            diff_stats[f] = None
            continue

        best_dir = v[0][1]

        if len(data[k]) == 1 or len(set([t[0] for t in v])) == 0:
            best_stats[f] = None
            if k == 'g_solved':
                x = sorted(\
                    [(data['time_cpu'][d][f], d) \
                        for d in g_args.dirs \
                            if data['time_cpu'][d][f] is not None])
                if len(set([t[0] for t in x])) > 1:
                    best_stats[f] = x[0][1]
        else:
            best_stats[f] = best_dir

        if best_stats[f] is None or (len(v) > 1 and not f_cmp(v[0][0], v[1][0])):
            diff_stats[f] = None
        else:
            diff_stats[f] = best_dir

    assert(diff_stats.keys() == best_stats.keys())
    return diff_stats, best_stats


def _format_field(field, width, color=None, colspan=0, classes=[]):
    field = "-" if field is None \
                            or (g_args.vbsd \
                                and not isinstance(field, str) \
                                and color != COLOR_DIFF \
                                and color != COLOR_BEST) \
                else str(field)
    if g_args.html:
        tag = "td"
        if color is not None and color != COLOR_NOCOLOR:
            classes.append(HTML_CLASS[color])
        if "header" in classes:
            tag = "th"
            classes = classes[:]
            classes.remove("header")
        c = ' class="{}"'.format(" ".join(classes)) if len(classes) > 0 else ''
        colspan = ' colspan="{}"'.format(colspan) if colspan > 0 else ''
        field = "<{0:s}{1:s}{2:s}>{3:s}</{0:s}>".format(tag, c, colspan, field)
    elif g_args.csv:
        if field == '-':
            return ''
        return field
    else:
        field = field.rjust(width)
        if color:
            field = "{}{}{}".format(color, field, COLOR_NOCOLOR)

    return field


def _print_row (columns, widths, colors=None, colspans=None, classes=[]):
    assert(len(columns) == len(widths))

    formatted_cols = []
    for i in range(len(columns)):
        column = columns[i]
        width = widths[i]
        color = None if colors is None else colors[i]
        colspan = 0 if colspans is None else colspans[i]
        clazzes = [] if len(classes) == 0 else classes[i]
        if isinstance(column, list):
            assert(isinstance(width, list))
            assert(isinstance(clazzes, list))
            cols = []
            for j in range(len(column)):
                clz = [] if len(clazzes) == 0 else clazzes[j]
                cols.append(_format_field(column[j], width[j], color,
                                          colspan=colspan, classes=clz))
            if g_args.csv:
                column = CSV_DELIM.join(cols)
            else:
                column = "".join(cols)
        else:
            column = _format_field(column, width, color, colspan, clazzes)

        formatted_cols.append(column)

    if g_args.plain:
        print("{}".format(" ".join(formatted_cols)))
    elif g_args.html:
        print("<tr>{}</tr>".format("".join(formatted_cols)))
    elif g_args.csv:
        print(CSV_DELIM.join(formatted_cols))
    else:
        print("{} |".format(" | ".join(formatted_cols)))


def _print_html_header():

    style_css = """
        * { font-size: 11px; color: #555; font-family: Helvetica, sans-serif; }
        table {
            border-spacing: 0;
            border-collapse: collapse;
            border: 2px solid #555;
        }
        th {
          padding: 5px;
          font-weight: bold;
          background-color: #555;
          color: #fff;
          text-align: right;
        }
        td { padding: 2px 4px; text-align: right; }
        tr:nth-child(2n) { background-color: #efefef; }
        .borderleft { border-left: 2px solid #555; }
        .nowrap { white-space: nowrap; }
        td.best { color: blue; }
        td.diff { font-weight: bold; color: blue; }
        td.disc { font-weight: bold; background-color: red; color: #fff; }
        td.stat { background-color: brown; color: #fff; }
    """

    print("""<html>
               <head>
                 <style>{}</style>
               </head>
               <body>""".format(style_css))


def _print_html_footer():
    print("<tbody></table></body></html>")


def _get_column_name(key):
    global g_args
    if g_args.csv:
        return key
    if key in filter_log_dict:
        return filter_log_dict[key][0]
    elif key in filter_err_dict:
        return filter_err_dict[key][0]

    assert(key in GROUP_COLUMNS)
    return GROUP_COLUMNS[key]

def _get_color(f, d, diff_stats, best_stats):
    global g_args
    if f in g_file_stats['status'][d] \
       and g_file_stats['status'][d][f] == STATUS_TIME:
        return COLOR_STAT
    if diff_stats[f] == d:
        return COLOR_DIFF
    elif best_stats[f] == d:
        return COLOR_BEST
    return COLOR_NOCOLOR


def _format_results(stat, values):
    global g_format_column_sum

    if len(values) == 0:
        return None

    if stat in g_format_column_sum:
        val = g_format_column_sum[stat](values)
    else:
        try:
            val = sum(values)
        except TypeError:
            val = None
        if isinstance(val, float):
            val = round(val, 1)
    return val

def _get_family_name(len_prefix, f):
    global g_args, g_file_stats
    f = g_file_stats['benchmark'][g_args.dirs[0]][f][len_prefix:]
    if g_args.grouped_families:
        end = f.find('/')
    else:
        end = f.rfind('/')
    if end > -1:
        return f[:end]
    else:
        return f.split('-')[0]

def _group_results():
    global g_args, g_benchmarks, g_file_stats
    stats = {}
    totals = {}

    stats[TOTALS_KEY] = {}

    len_prefix = len(os.path.commonprefix(
      list(g_file_stats['benchmark'][g_args.dirs[0]].values())))

    # collect statistics per group
    for f in g_benchmarks:
        group_name = _get_family_name(len_prefix, f)
        if group_name not in stats:
            stats[group_name] = {}

        group = stats[group_name]
        for d in g_args.dirs:
            if d not in group:
                group[d] = {}
            if d not in stats[TOTALS_KEY]:
                stats[TOTALS_KEY][d] = {}

            for stat in g_file_stats:
                if stat not in group[d]:
                    group[d][stat] = []
                if stat not in stats[TOTALS_KEY][d]:
                    stats[TOTALS_KEY][d][stat] = []
                val = g_file_stats[stat][d][f]
                # val is None if file does not exist or all files in
                # a group timed out for virtual best solver
                if val is not None:
                    group[d][stat].append(val)
                    stats[TOTALS_KEY][d][stat].append(val)

    # compute group totals
    for k in g_file_stats:
        assert(k not in totals)
        totals[k] = {}
        for d in g_args.dirs:
            if d not in totals[k]:
                totals[k][d] = {}

            for group in stats:
                assert(group not in totals[k][d])
                totals[k][d][group] = _format_results(k, stats[group][d][k])

    return totals, stats.keys()


def _base_dir(path):
    return os.path.basename(path.rstrip('/'))


def _print_totals():

    data, benchmarks = _group_results()

    benchmark_column_width, header_column_widths, data_column_widths = \
        _get_column_widths(data, g_args.dirs)

    # only some of the columns are interesting for the totals
    columns = ['g_total', 'g_solved']
    if not g_args.synth:
      columns += ['g_sat', 'g_unsat']
    if g_args.unique:
      columns += ['g_uniq']
    columns += ['g_unknown', 'g_time', 'g_mem', 'g_err', 'time_cpu', 'memory']
    if not g_args.html:
        print("\n")

    # header
    cols = ["DIRECTORY"]
    cols.extend(_get_column_name(k) for k in columns)
    widths = [max(len(_base_dir(d)) for d in g_args.dirs)]
    widths.extend([max(data_column_widths[k][d] for d in g_args.dirs) \
                      for k in columns])
    classes = [["header"] for i in range(len(columns) + 1)]
    _print_row(cols, widths, classes=classes)

    # totals
    for d in g_args.dirs:
        cols = [_base_dir(d)]
        cols.extend(data[k][d][TOTALS_KEY] for k in columns)
        _print_row(cols, widths)#, classes=classes)
    if g_args.html:
        print("</table><br><br><table>")
    else:
        print("\n\n")

def _get_column_widths(data, benchmarks):
    global g_args

    columns_all = [k for k in STATS_KEYS if k in data.keys()]
    columns_print = [k for k in g_args.columns if k in columns_all]

    padding = 1
    benchmark_column_width = \
        padding + max(max(len(b) for b in benchmarks), len("DIRECTORY"))

    data_column_widths = dict((k, {}) for k in columns_all)
    for d in g_args.dirs:
        for k in columns_all:
            data_column_widths[k][d] = \
                padding + \
                max(len(_get_column_name(k)),
                    max(len(str(v)) for v in data[k][d].values()))


    # header column widths
    column_groups = dict((d, list(columns_print)) for d in g_args.dirs)
    header_column_widths = {}
    for d, g in column_groups.items():
        width_cols = 0
        for k in g:
            width_cols += data_column_widths[k][d]
        width_dir = len(_base_dir(d))
        if width_cols > width_dir or g_args.plain:
            header_column_widths[d] = width_cols
        else:
            header_column_widths[d] = width_dir
            # add width to first column of column group if directory name is
            # longer than the column group itself
            data_column_widths[columns_print[0]][d] += width_dir - width_cols

    return benchmark_column_width, header_column_widths, data_column_widths


def _print_data ():
    global g_file_stats

    data = g_file_stats
    benchmarks = g_benchmarks

    # group results by benchmark group/family
    if g_args.grouped:
        data, benchmarks = _group_results()

    diff_stats, best_stats = _pick_data(benchmarks, data)

    benchmark_column_width, header_column_widths, data_column_widths = \
        _get_column_widths(data, benchmarks)

    if g_args.html:
        _print_html_header()
        print("""<table id="legend">
                   <tr>
                     <th>LEGEND</th>
                     <td class="best">best value</td>
                     <td class="diff">difference of '{}' >= {}</td>
                     <td class="disc">discrepancy</td>
                     <td class="stat">status differs</td>
                   </tr>
                 </table><br><br><table>""".format(g_args.cmp_col, g_args.dmin))

    if g_args.totals:
        _print_totals()

    if g_args.nocomp:
      if g_args.html:
        _print_html_footer()
      return

    # print header
    if not g_args.csv or len(g_args.dirs) > 1:
        columns = ["DIRECTORY"]
        columns.extend([_base_dir(d) for d in g_args.dirs])
        widths = [benchmark_column_width]
        widths.extend([header_column_widths[d] for d in g_args.dirs])
        colspans = [0]
        colspans.extend([len(g_args.columns) for d in g_args.dirs])

        classes = [["header"]]
        classes.extend([["borderleft", "header"] for d in g_args.dirs])
        if not g_args.plain:
            _print_row (columns, widths, colspans=colspans, classes=classes)

    columns = ["Benchmark"]
    if g_args.csv:
        columns = ['benchmark']
    widths = [benchmark_column_width]
    classes = [["header"]]

    for d in g_args.dirs:
        classes.append([["header"] for k in g_args.columns])
        classes[-1][0].append("borderleft")
        columns.append([_get_column_name(k) for k in g_args.columns])
        widths.append([data_column_widths[k][d] for k in g_args.columns])
    if not g_args.plain:
        _print_row (columns, widths, classes=classes)

    if g_args.html:
        print("</thead><tbody>")

    # print data rows
    rows = sorted(benchmarks, key=lambda s: s.lower())
    # totals row should always be on the bottom
    if g_args.grouped:
        assert (TOTALS_KEY in rows)
        rows.remove(TOTALS_KEY)
        rows.append(TOTALS_KEY)

    for f in rows:
        assert 'status' in data
        s = _get_values(data, f, 'status')

        # highlight row if status is not the same or solvers are disagreeing
        color = COLOR_DISC if STATUS_DIS in s else COLOR_NOCOLOR

        columns = []
        if g_args.show_full_path:
            columns.append(data['benchmark'][g_args.dirs[0]][f])
        else:
            columns.append(f)
        widths = [benchmark_column_width]
        colors = [color]
        classes = [["nowrap"]]

        for d in g_args.dirs:
            classes.append([["borderleft"]])
            classes[-1].extend([[] for i in range(len(g_args.columns) - 1)])
            columns.append([data[k][d][f] for k in g_args.columns])
            widths.append([data_column_widths[k][d] for k in g_args.columns])
            colors.append(color if color != COLOR_NOCOLOR \
                        else _get_color(f, d, diff_stats, best_stats))

        _print_row(columns, widths, colors=colors, classes=classes)

    if g_args.html:
        _print_html_footer()


def _get_values(data, f, k):
    return [data[k][d][f] for d in g_args.dirs]

def _has_status(data, f, status_list):
    file_status = set(_get_values(data, f, 'status'))
    return len(set(status_list).intersection(file_status)) > 0

def _filter_benchmarks(data, filterfunc):
    global g_benchmarks
    for f in g_benchmarks.copy():
        if not filterfunc(data, f):
            g_benchmarks.remove(f)

def _parse_arguments():
    global g_args
    aparser = ArgumentParser(
                  formatter_class=ArgumentDefaultsHelpFormatter,
                  epilog="availabe values for column: {{ {} }}, ".format(
                      ", ".join(sorted(STATS_KEYS))))
    aparser.add_argument \
          (
            "-f",
            metavar="string", dest="filter", type=str, default=None,
            action='append',
            help="Filter benchmark files by <string>"
          )
    aparser.add_argument \
          (
            "--fregex",
            metavar="string", dest="filter_regex", type=str, default=None,
            help="Filter benchmark files by <regex>"
          )
    aparser.add_argument \
          (
            "-e",
            metavar="string", dest="exclude", type=str, default=None,
            action="append",
            help="Exclude benchmark files that contain <string>"
          )
    aparser.add_argument \
          (
            "--to",
            metavar="seconds[,second ...]", dest="timeout", default=None,
            help="Use individual time out"
          )
    aparser.add_argument \
          (
            "--vb",
            action="store_true",
            dest='virtual_best',
            help="Generate virtual best solver results"
          )
    aparser.add_argument \
          (
            "--vbp",
            action="store_true",
            dest='virtual_best_pp',
            help="Generate virtual best solver out of two runs with " \
                 "one treated as a preprocessing step (don't forget to " \
                 "provide a list " \
                 "of timeouts via --to in the same order as the runs, " \
                 "the run with the lesser timeout serves as " \
                 "preprocessing step)"
          )
    aparser.add_argument \
          (
            "--vb-show-details",
            dest="vbsd", action="store_true",
            help="Show detailed overview for virtual best solver"
           )
    aparser.add_argument \
          (
            "--high-diff",
            metavar="float", dest="dmin", type=float, default=0.1,
            help="Highlight difference if greater than <float>"
          )
    aparser.add_argument \
          (
            "--discrepancy",
            action="store_true",
            dest='show_discrepancy',
            help="Show discrepancies only"
          )
    aparser.add_argument \
          (
            "--ignore-dis",
            action="store_true",
            dest='ignore_discrepancy',
            help="Ignore discrepancies"
          )
    aparser.add_argument \
          (
            "--time",
            action="store_true",
            dest='show_time',
            help="Show timeouts only"
          )
    aparser.add_argument \
          (
            "--toof",
            metavar="dir", dest="timeof", action="store", default=None,
            help="Show timeouts of given dir only"
          )
    aparser.add_argument \
          (
            "--mem",
            action="store_true",
            dest='show_mem',
            help="Show memory outs only"
          )
    aparser.add_argument \
          (
            "--err",
            action="store_true",
            dest='show_err',
            help="Show errors only"
          )
    aparser.add_argument \
          (
            "--ok",
            action="store_true",
            dest='show_ok',
            help="Show non-errors only"
          )
    aparser.add_argument \
          (
            "-c", "--compare",
            metavar="column", dest="cmp_col",
            choices=STATS_KEYS,
            help="Compare results column"
          )
    aparser.add_argument \
          (
            "-s",
            metavar="column[,column ...]", dest="columns",
            help="List of columns to print"
          )
    aparser.add_argument \
          (
            "-a",
            dest="show_all", action="store_true",
            help="Print all columns"
          )
    aparser.add_argument \
          (
            "--html",
            action="store_true",
            dest='html',
            help="Generte html output"
          )
    aparser.add_argument \
          (
            "--csv",
            action="store_true",
            dest='csv',
            help="Generate csv output"
          )
    aparser.add_argument \
          (
            "dirs",
            nargs='+',
            help="Solver runs to compate"
          )
    aparser.add_argument \
          (
            "--no-colors",
            action="store_true",
            dest='nocolors',
            default=False,
            help="Disable colored output"
          )
    aparser.add_argument \
          (
            "--no-comp",
            action="store_true",
            dest='nocomp',
            default=False,
            help="Do not print comparison"
          )
    aparser.add_argument \
          (
            "-g", "--grouped",
            dest="grouped",
            action="store_true",
            help="Group benchmarks into sub-families"
          )
    aparser.add_argument \
          (
            "-G", "--grouped-families",
            dest="grouped_families",
            action="store_true",
            help="Group benchmarks into families"
          )
    aparser.add_argument \
          (
            "-u", "--unique",
            action="store_true",
            dest='unique',
            help="Show column with uniquely solved instances"
          )
    aparser.add_argument \
          (
            "-t", "--total",
            dest="totals",
            action="store_true",
            help="Show totals table"
          )
    aparser.add_argument \
        (
            "--synth",
            dest="synth", default=False, action="store_true",
            help="Use presence of solutions as criterion for solved"
        )
    aparser.add_argument \
        (
            "--tptp",
            dest="tptp", default=False, action="store_true",
            help="Parse SAT/UNSAT in terms of TPTP output"
        )
    aparser.add_argument \
          (
            "-p", "--plain",
            dest="plain", action="store_true",
            help="Plain mode, only show columns and " \
                 "filename (for gnuplot data files)"
          )
    aparser.add_argument \
          (
            "--common",
            dest="show_common", default=False, action="store_true",
            help="Show commonly solved instances only"
          )
    aparser.add_argument \
          (
            "--unsolved",
            dest="show_unsolved", default=False, action="store_true",
            help="Show unsolved instances only"
          )
    aparser.add_argument \
          (
            "-d", "--diff",
            dest="show_diff", default=False, action="store_true",
            help="Show differences only (see --diff-column)"
          )
    aparser.add_argument \
          (
            "--diff-column",
            metavar="column", dest="diff_col",
            choices=STATS_KEYS,
            default="status",
            help="Column to compare for differences (see --diff)"
          )
    aparser.add_argument \
          (
            "--full-path",
            dest="show_full_path", default=False, action="store_true",
            help="Show full path instead of benchmark name"
          )
    aparser.add_argument \
          (
            "--penalty", type=int,
            metavar="seconds[,second ...]", dest="pen", default=None,
            help="CPU time penalty for memory out/error"
          )
    aparser.add_argument \
        (
            "--sat",
            dest="show_sat", default=False, action="store_true",
            help="Show instances with result 'sat'"
        )
    aparser.add_argument \
        (
            "--no-sat",
            dest="show_no_sat", default=False, action="store_true",
            help="Don't show instances with result 'sat'"
        )
    aparser.add_argument \
        (
            "--unsat",
            dest="show_unsat", default=False, action="store_true",
            help="Show instances with result 'unsat'"
        )
    aparser.add_argument \
        (
            "--no-unsat",
            dest="show_no_unsat", default=False, action="store_true",
            help="Don't show instances with result 'unsat'"
        )
    aparser.add_argument \
        (
            "--unknown",
            dest="show_unknown", default=False, action="store_true",
            help="Show instances with result 'unknown'"
        )
    aparser.add_argument \
        (
            "-i", "--incremental",
            dest="incremental", default=False, action="store_true",
            help="Compare incremental runs"
        )
    g_args = aparser.parse_args()

    if g_args.grouped_families:
        g_args.grouped = True


if __name__ == "__main__":
    try:
        _parse_arguments()

        # do not use a set here as the order of directories should be preserved
        unique_dirs = []
        for d in g_args.dirs:
            if d not in unique_dirs:
                unique_dirs.append(d)
        g_args.dirs = unique_dirs

        if len(g_args.dirs) < 1:
            raise CmpSMTException ("No directories to compare specified")

        if g_args.cmp_col is None:
            if g_args.grouped:
                g_args.cmp_col = 'g_solved'
            else:
                g_args.cmp_col = 'time_cpu'

        if g_args.virtual_best_pp: g_args.virtual_best = True

        for d in g_args.dirs:
            if not os.path.isdir(d):
                raise CmpSMTException (
                        "Given directory '{}' does not exist".format(d))

        if g_args.columns is None:
            if g_args.grouped:
                g_args.columns = \
                    "status,result,g_solved,g_total,g_time,g_mem,g_err," \
                    "time_cpu,memory"
            else:
                g_args.columns = "status,result,exit_code,time_cpu,memory"

        g_args.columns = g_args.columns.split(',')
        for c in g_args.columns:
            if c not in STATS_KEYS:
                raise CmpSMTException("column '{}' not available".format(c))

        if g_args.show_all:
            g_args.columns = STATS_KEYS

        if g_args.timeout:
            g_args.timeout = [float(s) for s in g_args.timeout.split(',')]
            if len(g_args.timeout) > 1 \
               and len(g_args.timeout) != len(g_args.dirs):
                   raise CmpSMTException (
                           "Invalid number of timeouts given." \
                           "Number of timeouts must match number of " \
                           "directories")
            if g_args.timeout and len(g_args.timeout) == 1:
                g_args.timeout = [g_args.timeout[0] for d in g_args.dirs]
            assert (len(g_args.timeout) == len(g_args.dirs))
            tmp = { g_args.dirs[i]:g_args.timeout[i] \
                    for i in range(0, len(g_args.dirs)) }
            g_args.timeout = tmp

        # some columns may not make sense or are not available for some options
        remove_columns = []
        if g_args.grouped and 'result' in g_args.columns:
            remove_columns.append('result')

        for c in remove_columns:
            if g_args.columns.count(c) > 0:
                g_args.columns.remove(c)

        if g_args.unique:
            g_args.columns.append('g_uniq')

        if g_args.incremental:
            g_args.columns = ['status', 'g_solved', 'g_sat', 'g_unsat',
                              'time_cpu', 'memory']

        if g_args.csv:
            g_args.show_full_path = True

        if len(g_args.columns) == 0:
            raise CmpSMTException ("No columns selected to print")

        if g_args.nocolors:
            COLOR_BEST = ''
            COLOR_DIFF = ''
            COLOR_DISC = ''
            COLOR_STAT = ''
            COLOR_NOCOLOR = ''

        if g_args.timeof and not g_args.timeof in g_args.dirs:
            raise CmpSMTException (
                    "Invalid directory '{}' given".format(g_args.timeof))

        _read_data(g_args.dirs)

        # remove files not to display
        if g_args.filter_regex:
            pattern = re.compile(g_args.filter_regex)
            filter = lambda d, f:  pattern.match(str(f))
            _filter_benchmarks(g_file_stats, filter)

        if g_args.filter:
            filter = lambda d, f: all(x in str(f) for x in g_args.filter)
            _filter_benchmarks(g_file_stats, filter)

        if g_args.exclude:
            exclude = lambda d, f: not any(x in str(f) for x in g_args.exclude)
            _filter_benchmarks(g_file_stats, exclude)

        if len(g_file_stats.keys()) > 0:
            _init_missing_files (g_file_stats)
            _normalize_data(g_file_stats)

            if g_args.show_common:
                filter = \
                    lambda d, f: \
                        len(set(_get_values(d, f, 'status'))) == 1 and \
                        len(set(_get_values(d, f, 'result'))) == 1
                _filter_benchmarks(g_file_stats, filter)
            if g_args.show_unsolved:
                filter = lambda d, f: \
                        len(set(_get_values(d, f, 'result')).intersection(\
                            {RESULT_SAT, RESULT_UNSAT, RESULT_SYNTH})) == 0
                _filter_benchmarks(g_file_stats, filter)
            if g_args.show_diff:
                filter = lambda d, f: \
                        len(set(_get_values(d, f, g_args.diff_col))) != 1
                _filter_benchmarks(g_file_stats, filter)

            result = None
            if g_args.show_sat:
                result = RESULT_SAT
            if g_args.show_unsat:
                result = RESULT_UNSAT
            # FIXME: --unknown not working as expected
            if g_args.show_unknown:
                result = RESULT_UNKNOWN
            if result and not g_args.incremental:
                filter = lambda d,f: result in set(_get_values(d, f, 'result'))
                _filter_benchmarks(g_file_stats, filter)

            if g_args.show_no_sat:
                filter = \
                    lambda d,f: \
                        RESULT_SAT not in set(_get_values(d, f, 'result'))
                _filter_benchmarks(g_file_stats, filter)

            if g_args.show_no_unsat:
                filter = \
                    lambda d,f: \
                        RESULT_UNSAT not in set(_get_values(d, f, 'result'))
                _filter_benchmarks(g_file_stats, filter)

            filter_status = []
            if g_args.show_err:
                filter_status.append(STATUS_ERR)
            if g_args.show_mem:
                filter_status.append(STATUS_MEM)
            if g_args.show_ok:
                filter_status.append(STATUS_OK)
            if g_args.show_time:
                filter_status.append(STATUS_TIME)
            if g_args.show_discrepancy:
                filter_status.append(STATUS_DIS)
            if filter_status:
                filter = lambda data, f: _has_status(data, f, filter_status)
                _filter_benchmarks(g_file_stats, filter)

            if g_args.timeof:
                d = g_args.timeof
                filter = lambda data, f: data['status'][d][f] == STATUS_TIME
                _filter_benchmarks(g_file_stats, filter)

            if len(g_benchmarks) == 0:
                print("No files found that match given criteria")
            else:
                _print_data ()
        else:
            if g_args.filter:
                print("No files found that match '{}'".format(g_args.filter))
            else:
                print("No files found")

    except KeyboardInterrupt as e:
        sys.exit ("[cmpsmt] interrupted")
    except CmpSMTException as e:
        sys.exit (str(e))
    # ignore broken pipe errors (can occur if compare-smt-runs.py | head -n1 is
    # called)
    except BrokenPipeError:
        pass
