Synthesis Artifact Description
=======================================

# Table of Contents

1. [Directory Structure](#directory-structure)
2. [Workflow Overview](#overview-of-the-synthesis-workflow)
3. [Workflow Inputs](#input-to-the-workflow)
4. [Files generated by Workflow](#files-generated-by-the-workflow)
5. [Synthesis Example](#synthesis-example)


# Directory Structure

## Directories

  * `grammars`: Contains custom grammars for some synthesis tasks, which can
    be used instead of the provided grammar templates.

  * `scripts`: Contains scripts to generate the initial directory structure.

  * `example`: Contains the example discussed in [Synthesis Example](#synthesis-example).


  * `full_io_spec`:
    Contains all files required for synthsizing invertibility conditions for a
    given floating-point operator/predicate/rounding mode combination for a
    given floating-point format.
    The subdirectories are structured as follows:

    ```
      full_io_spec/<FORMAT>/<ROUNDING_MODE>/<OPERATOR>/<PREDICATE>/
    ```

    Each final subdirectory contains:

      * `final_solution.smt2`: Pre-computed final synthesized solution.
      * `full_io_spec.smt2`: Pre-computed I/O specification.
      * `input_grammar.sy`:  Input grammar, see [Workflow Inputs](#input-to-the-workflow).
      * `input_problem.smt2`: Input problem, see [Workflow Inputs](#input-to-the-workflow).

    **[NOTE]**
    * We added the pre-computed files `final_solution.smt2` and
      `full_io_spec.smt2` for time saving reasons since computing them is in
      many cases very time consuming (as pointed out in the paper).
    * For the cases for which we were not able to synthesize an invertibility
      condition the `final_solution.smt2` file will be missing.
    * For the rounding mode independent operators `fp.abs`, `fp.neg`, and
      `fp.rem` the files are the same for all rounding modes. Thus, we only
      keep the RNE subdirectories.



# Overview of the Synthesis Workflow

To use the workflow, at a high level, consists of the following strategy,
given in terms of scripts to run.

Refer to the last section for a concrete example.

1. **Setup**

  1. Initialize `input_problem.smt2` and `input_grammar.sy` (see [Workflow Inputs](#input-to-the-workflow)).

  2. If needed, clean the working directory.

    ```
      run_gen_ic -clean

    ```

  3. Cache the I/O specification of the invertibility condition.

    This step can take up to 2 hours for two dimensional invertibility conditions
    for `(_ FloatingPoint 3 5)`, and can be shorter or longer depending on the
    operator.

    **[NOTE]** For convenience, the artifact comes with all precomputed I/O
    specifications.

    ```
      run_gen_ic -setup [--gen-ic-use-eval]

    ```

    Option `--gen-ic-use-eval` must be used if the satisfiability problem
    for `input_problem.smt2` for fixed values of variables is hard.
    For example, this is the case if `input_problem.smt2` involves FP
    multiplication.

    **[OPTIONAL]**
    Visualize the specification. This only is intended to work
    for two dimensional invertibility conditions, i.e., invertibility
    conditions involving a binary FP operator and a binary FP predicate
    (e.g., `fp.add` with `fp.gt`).

    ```
      run_gen_ic -visual full_io_spec.smt2 > image.ppm
    ```

    You can view the image with an image viewer that either supports the ppm
    format or you can convert the image to JPG with:

    ```
      convert image.ppm image.jpg
    ```

2. **Synthesis Loop**

  1. Decide on a part of the search by modifying the side condition in
     `input_problem.smt2`, by default this is `true`.

  2. Initialize the synthesis environment.

    ```
      run_gen_ic -reset
    ```

  3. Run alternating calls to synthesize/test new candidates.

    1. Synthesize a new candidate.
       Decide on an input grammar, based on the perceived fitness of terms, by
       modifying `input_grammar.sy`.
       Synthesize a list of solutions (in decendending order of size).

        ```
          run_gen_ic -synth
        ```

       **[NOTE]** The user must kill the above process when a reasonable
       candidate is generated.

    2. Test the current candidate until up to `NUM_PBE_PTS` counter examples.
      ```
        run_gen_ic -test
      ```

        + If the test succeeded, test the current candidate for the entire
          specification.

          ```
            run_gen_ic -testfull
          ```

          If verified, we are finished and solutions.smt2 is appended.

        + If the test failed you can optionally cache, evaluate, and visualize
          the current candidate.

          ```
            run_gen_ic -eval    # Evaluate and generate I/O specification of
                                # current candidate
          ```
          ```
            run_gen_ic -gen [CANDIDATE_FILE].smt2 > [IO_OUT].smt2;
            run_gen_ic -visual [IO_OUT].smt2 >  [IMAGE_OUT].ppm
          ```

          Othwerwise, continue synthesis and repeat step iii.

  The above synthesis loop can also be automated via `run_gen_ic -run`.
  However, notice that this option will not necessarily lead to minimal
  solutions, nor is it the fastest way to find a solution. In practice, one
  needs to allocate a custom amount of time to `run_gen_ic -synth` to wait
  for "good" solutions.


The following describes the necessary files to set up for this workflow, and
the intermediate files that are generated, and how they can be modified.


Input to the Workflow
===============================================================================

* `input_problem.smt2`:

  The definition of the invertibility condition problem. This is a template
  in SMT-LIBv2 format that corresponds to an invertibility condition problem.


* `input_grammar.sy`:

  The input grammar for solving the invertibility condition. This
  is a SyGuS definition of a synth-fun command for predicate `IC`.
  The type of `IC` should match the type of `input` from `input_problem.smt2`.
  This grammar should be focused on the invertibility condition problem. In
  other words, it contains the guess of a class of expressions that contains
  the invertibility condition.



# Files generated by the Workflow

* `full_io_spec.smt2`:

  This is a cache of the I/O behavior of the input problem. This is a
  definition of an array `io` whose constituent type is `BitVector`. Each bit
  in the range of this array corresponds to a point of the invertibility
  conditions behavior, where 0 means it holds, and 1 means it does not.


* `pbe_example_set.sy`:

  This corresponds to the approximation of the problem we are currently
  considering. This contains, in SyGuS format, constraints of the form:

  ```
    (constraint (IC c1 ... cn))
    ...
    (constraint (not (IC d1 ... dn)))
  ```

  constants c1, ...,cn and d1,...,dn. These points are updated by `run_gen_ic -reset`,
  `run_gen_ic -test` and `run_gen_ic -testfull`. In all cases, the points
  appended to this list are counterexamples to failed candidates. The number of
  points generated by these calls can be configured in `run_gen_ic` via the
  variable `NUM_PBE_PTS`. Additionally, all points generated by the first two
  satisfy the side condition from `input_problem.smt2`. This set can be
  modified as necessary if the user has an idea of which points are useful or
  not useful (although this is not common).


* `input_candidates.smt2`:

  This is the current list of candidate invertibility conditions generated by
  the synthesis component of the workflow. This is, in SMT-LIBv2 format, a list
  of definitions `IC`. The last definition in this list is used as the
  candidate invertibility condition when running `run_gen_ic -test` or
  `run_gen_ic -testfull`. The user can modify this file with a custom `IC` when
  they have a manual guess for the next candidate.


* `candidate_log.smt2`:

  This is a list of candidate invertibility conditions that have been tested
  since the last call to reset. In other words, these are the input conditions
  that have been extracted by `run_gen_ic -test`, where `run_gen_ic -reset`
  clears this list.


* `candidate_log_overall.smt2`:

  A concatenation of `candidate_log.smt2` generated during the workflow. This
  is updated by `run_gen_ic -reset`.


* `solutions.smt2`:

  After a successful call to `run_gen_ic -testfull`, this is updated to contain
  the solution of the invertibility condition problem.



# Synthesis Example

  In the following, we use fp.abs with equality as a concrete example
  for our synthesis workflow.

  The `example` directory contains all input files necessary to run the
  synthesis loop and corresponds to directory `full_io_spec/3_5/RNE/fp.abs/\=/`.

  ```
  $ cd example

  $ ../../run_gen_ic -reset
  *** Synthesizing first initial candidate...
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool true)
  ***** Testing first initial candidate:
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool true)

  $ ../../run_gen_ic -run
  *** Synthesizing first initial candidate...
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool true)
  ***** Testing first initial candidate:
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool true)

  ***** Generated new counterexamples:
  (constraint (not (IC (fp #b1 #b101 #b1111) )))
  (constraint (not (IC (fp #b1 #b110 #b0111) )))
  (constraint (not (IC (fp #b1 #b000 #b0100) )))
  (constraint (not (IC (fp #b1 #b110 #b1100) )))
  (constraint (not (IC (fp #b1 #b010 #b0001) )))
  (constraint (not (IC (fp #b1 #b110 #b0100) )))
  (constraint (not (IC (fp #b1 #b110 #b0000) )))
  (constraint (not (IC (fp #b1 #b110 #b0011) )))
  (constraint (not (IC (fp #b1 #b100 #b1111) )))
  (constraint (not (IC (fp #b1 #b100 #b1001) )))
  *** Synthesizing second initial candidate...
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool false)
  ***** Testing second initial candidate:
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool false)

  ***** Generated new counterexamples:
  (constraint (IC (fp #b0 #b100 #b0010) ))
  (constraint (IC (fp #b0 #b010 #b1111) ))
  (constraint (IC (fp #b0 #b010 #b0111) ))
  (constraint (IC (fp #b0 #b110 #b1001) ))
  (constraint (IC (fp #b0 #b000 #b1001) ))
  (constraint (IC (fp #b0 #b000 #b1011) ))
  (constraint (IC (fp #b0 #b110 #b0001) ))
  (constraint (IC (fp #b0 #b011 #b0110) ))
  (constraint (IC (fp #b0 #b011 #b1000) ))
  (constraint (IC (fp #b0 #b010 #b0110) ))
  *** Synthesizing candidates...
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool (ite (fp.isNegative t) false true))
  ***** Testing candidate:
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool (ite (fp.isNegative t) false true))

  ***** No counterexamples for candidate!!!

   $ ../../run_gen_ic -simplify input_candidates.smt2
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool (not (fp.isNegative t)))
  ```

  Alternatively, this loop can also be performed manually:

  ```
  $ cd example

  $ ../../run_gen_ic -reset
  *** Synthesizing first initial candidate...
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool true)
  ***** Testing first initial candidate:
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool true)

  ***** Generated new counterexamples:
  (constraint (not (IC (fp #b1 #b101 #b1111) )))
  (constraint (not (IC (fp #b1 #b110 #b0111) )))
  (constraint (not (IC (fp #b1 #b000 #b0100) )))
  (constraint (not (IC (fp #b1 #b110 #b1100) )))
  (constraint (not (IC (fp #b1 #b010 #b0001) )))
  (constraint (not (IC (fp #b1 #b110 #b0100) )))
  (constraint (not (IC (fp #b1 #b110 #b0000) )))
  (constraint (not (IC (fp #b1 #b110 #b0011) )))
  (constraint (not (IC (fp #b1 #b100 #b1111) )))
  (constraint (not (IC (fp #b1 #b100 #b1001) )))
  *** Synthesizing second initial candidate...
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool false)
  ***** Testing second initial candidate:
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool false)

  ***** Generated new counterexamples:
  (constraint (IC (fp #b0 #b100 #b0010) ))
  (constraint (IC (fp #b0 #b010 #b1111) ))
  (constraint (IC (fp #b0 #b010 #b0111) ))
  (constraint (IC (fp #b0 #b110 #b1001) ))
  (constraint (IC (fp #b0 #b000 #b1001) ))
  (constraint (IC (fp #b0 #b000 #b1011) ))
  (constraint (IC (fp #b0 #b110 #b0001) ))
  (constraint (IC (fp #b0 #b011 #b0110) ))
  (constraint (IC (fp #b0 #b011 #b1000) ))
  (constraint (IC (fp #b0 #b010 #b0110) ))

  *** Synthesis environment initialized. Now, run alternating calls to:
    run_gen_ic -synth
    run_gen_ic -test

  $ ../../run_gen_ic -synth
  *** Synthesizing candidates...
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool (ite (fp.isNegative t) false true))
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool (fp.isPositive t))
  ^C  # Note: interrupt with Ctrl-C

  $ ../../run_gen_ic -test
  ***** Testing candidate:
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool (fp.isPositive t))

  ***** Generated new counterexamples:
  (constraint (IC (fp #b0 #b111 #b1000) ))

  $ ../../run_gen_ic -synth
  *** Synthesizing candidates...
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool (ite (fp.isNegative t) false true))
  ^C  # Note: interrupt with Ctrl-C

  $ ../../run_gen_ic -test
  ***** Testing candidate:
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool (ite (fp.isNegative t) false true))

  ***** No counterexamples for candidate!!!

   $ ../../run_gen_ic -simplify input_candidates.smt2
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool (not (fp.isNegative t)))

  ```

  The output of the last command corresponds to the final solution provided in

  ```
  $ cat final_solution.smt2
  (define-fun IC ((t (_ FloatingPoint 3 5))) Bool (not (fp.isNegative t)))
  ```




