#!/usr/bin/env python

import argparse
import csv
import os
import sys


ROUNDING_MODES = ['RNE', 'RNA', 'RTP', 'RTN', 'RTZ']
ROUNDING_MODE_OPS = ['fp.add', 'fp.sub', 'fp.mul', 'fp.div', 'fp.fma',
                     'fp.sqrt', 'fp.roundToIntegral']
COMMUTATIVE_OPS = ['fp.add', 'fp.mul']


INPUT_PROBLEM = """(set-logic ALL)
(define-sort FP () (_ FloatingPoint {eb} {sb}))
(declare-fun IC ({ic_signature}) Bool)
(assert
 (forall ({forall_vars})
  (=>
   {side_condition}
   (=
    (IC {ic_args})
    (exists ((x FP))
     {problem}
    )
   )
  )
 )
)
"""

INPUT_PROBLEM_DIM0 = """(set-logic ALL)
(define-sort FP () (_ FloatingPoint {eb} {sb}))
(declare-fun IC ({ic_signature}) Bool)
(assert
 (=>
  {side_condition}
  (=
   IC
   (exists ((x FP))
    {problem}
   )
  )
 )
)
"""

SPECIAL_CONSTS = """\
(define-const posMinSubnormal FP (fp #b0 #b{e_zero} #b{s_one}))
(define-const negMinSubnormal FP (fp #b1 #b{e_zero} #b{s_one}))
(define-const posMaxSubnormal FP (fp #b0 #b{e_zero} #b{s_ones}))
(define-const negMaxSubnormal FP (fp #b1 #b{e_zero} #b{s_ones}))
(define-const posMinNormal FP (fp #b0 #b{e_one} #b{s_zero}))
(define-const negMinNormal FP (fp #b1 #b{e_one} #b{s_zero}))
(define-const posMaxNormal FP (fp #b0 #b{e_notone} #b{s_ones}))
(define-const negMaxNormal FP (fp #b1 #b{e_notone} #b{s_ones}))
(define-const posConstTwo FP (fp #b0 #b{e_two} #b{s_zero}))
(define-const R RoundingMode {rm})
"""

GRAMMAR_TEMPLATE_DIM0 = """\
(synth-fun IC () Bool
  ((Start Bool (
     true
     false
))))"""

GRAMMAR_TEMPLATE = """\
(synth-fun IC ({inputs}) Bool
  ((Start Bool (
     true
     false

     (= StartFP StartFP)
     (fp.eq StartFP StartFP)
     (fp.leq StartFP StartFP)
     (fp.lt StartFP StartFP)

     (fp.isNormal StartFP)
     (fp.isSubnormal StartFP)
     (fp.isZero StartFP)
     (fp.isInfinite StartFP)
     (fp.isNaN StartFP)
     (fp.isNegative StartFP)
     (fp.isPositive StartFP)

     (ite Start Start Start)
   ))
   (StartFP FP (
     {input_symbols}
     posMinSubnormal
     negMinSubnormal
     posMaxSubnormal
     negMaxSubnormal
     posMinNormal
     negMinNormal
     posMaxNormal
     negMaxNormal
     posConstTwo

     (fp.abs StartFP)
     (fp.neg StartFP)
     (fp.sqrt StartRM StartFP)
     (fp.roundToIntegral StartRM StartFP)

     (fp.add StartRM StartFP StartFP)
     (fp.sub StartRM StartFP StartFP)
     (fp.mul StartRM StartFP StartFP)
     (fp.div StartRM StartFP StartFP)
     (fp.fma StartRM StartFP StartFP StartFP)

     (fp.rem StartFP StartFP)
     (fp.min StartFP StartFP)
     (fp.max StartFP StartFP)
   ))
   (StartRM RoundingMode
     (RNE RNA RTZ RTP RTN))
))
"""


def rm_required(op):
    return op in ROUNDING_MODE_OPS

def die(msg):
    print('[error] gen-input-problems.py: {}'.format(msg))
    sys.exit(1)

def main():
    ap = argparse.ArgumentParser()
    ap.add_argument('--eb', default=3, type=int)
    ap.add_argument('--sb', default=5, type=int)
    ap.add_argument('--rm', default='RNE')
    ap.add_argument('relations')
    ap.add_argument('operations')
    ap.add_argument('outdir')

    args = ap.parse_args()

    if not os.path.isfile(args.relations):
        die("Relations file '{}' does not exist".format(args.relations))
    if not os.path.isfile(args.operations):
        die("Relations file '{}' does not exist".format(args.operations))
#    if os.path.exists(args.outdir):
#        die("Output directory '{}' already exists".format(args.outdir))
    if args.rm not in ROUNDING_MODES:
        die("Invalid rounding mode '{}' specified".format(args.rm))

    relations = []
    with open(args.relations) as rel_file:
        rel_reader = csv.reader(rel_file, delimiter=',')
        for rel in rel_reader:
            relations.append(rel)

    operations = []
    with open(args.operations) as op_file:
        op_reader = csv.reader(op_file, delimiter=',')
        for op in op_reader:
            operations.append(op)

    e_zero = '0' * args.eb
    e_one = '{}1'.format('0' * (args.eb - 1))
    e_notone = '{}0'.format('1' * (args.eb - 1))
    e_two = '1{}'.format('0' * (args.eb - 1))
    s_one = '{}1'.format('0'* (args.sb - 2))
    s_ones = '1' * (args.sb - 1)
    s_zero = '0' * (args.sb - 1)

    if not os.path.isdir(args.outdir):
        os.mkdir(args.outdir)
    for op in operations:
        op_name = op[0]
        op_arity = int(op[1])
        side_condition = 'true'
# TODO: do we need SCs for generating input problems?
#        if len(op) == 3:
#            op_sc = op[2]
        d = os.path.join(args.outdir, op_name)
        if not os.path.isdir(d):
            os.mkdir(d)
        rm = ''
        if rm_required(op_name):
            rm = ' {}'.format(args.rm)

        for rel in relations:
            rel_name = rel[0]
            rel_arity = int(rel[1])

            consts = []
            if op_arity >= 2:
                consts.append('s')
            if op_arity >= 3:
                consts.append('u')
            if rel_arity == 2:
                consts.append('t')

            ic_args = ' '.join(consts)
            ic_signature = ' '.join(['FP' for i in range(len(consts))])
            forall_vars = ' '.join(['({} FP)'.format(c) for c in consts])

            op_operands = []
            if op_arity >= 2:
                op_operands.append('s')
            if op_arity == 3:
                op_operands.append('u')

            if op_name in COMMUTATIVE_OPS or op_name == '':
                num_variants = 1
            else:
                num_variants = op_arity

            for pos in range(num_variants):

                # fp.fma is commutative for the addition part, skip redundant
                # case
                if op_name == 'fp.fma' and pos == 1:
                    continue

                operands = op_operands[:]
                operands.insert(pos, 'x')
                if op_name:
                    op_smt2 = '({}{} {})'.format(op_name, rm, ' '.join(operands))
                else:
                    assert len(operands) == 1
                    op_smt2 = operands[0]

                rel_operands = [op_smt2]
                if rel_arity == 2:
                    rel_operands.append('t')

                problem = '({} {})'.format(rel_name, ' '.join(rel_operands))

                if num_variants > 1:
                    subdir_name = '{}{}'.format(rel_name, pos)
                else:
                    subdir_name = rel_name
                out_dir = os.path.join(args.outdir, op_name, subdir_name)
                if not os.path.isdir(out_dir):
                    os.mkdir(out_dir)

                if consts:
                    #out_file = os.path.join(out_dir, 'input_problem.smt2')
                    #with open(out_file, 'w') as problem_file:
                    #    problem = \
                    #        INPUT_PROBLEM.format(
                    #            eb=args.eb,
                    #            sb=args.sb,
                    #            ic_args=ic_args,
                    #            ic_signature=ic_signature,
                    #            forall_vars=forall_vars,
                    #            side_condition=side_condition,
                    #            problem=problem
                    #            )
                    #    problem_file.write(problem)


                    inputs = ['({} FP)'.format(x) for x in consts]

                    special_consts = SPECIAL_CONSTS.format(eb=args.eb,
                                            sb=args.sb,
                                            e_zero=e_zero,
                                            e_one=e_one,
                                            e_notone=e_notone,
                                            e_two=e_two,
                                            s_one=s_one,
                                            s_ones=s_ones,
                                            s_zero=s_zero,
                                            rm=args.rm)
                    grammar = GRAMMAR_TEMPLATE.format(
                                    inputs=forall_vars,
                                    input_symbols=' '.join(consts))
                else:
                    #out_file = os.path.join(out_dir, 'input_problem.smt2')
                    #with open(out_file, 'w') as problem_file:
                    #    problem = \
                    #        INPUT_PROBLEM_DIM0.format(
                    #            eb=args.eb,
                    #            sb=args.sb,
                    #            ic_signature=ic_signature,
                    #            forall_vars=forall_vars,
                    #            side_condition=side_condition,
                    #            problem=problem
                    #            )
                    #    problem_file.write(problem)

                    special_consts = ''
                    grammar = GRAMMAR_TEMPLATE_DIM0


                out_file = os.path.join(out_dir, 'input_grammar.sy')
                with open(out_file, 'w') as file:
                    file.write('{}{}'.format(special_consts, grammar))




if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        print()
        sys.exit('Interrupted')
